# __author__ = "Daniele Antonucci, Ulrich Filippi Oberegger, Olga Somova"
# __credits__ = ["Daniele Antonucci", "Ulrich FIlippi Oberegger", "Olga Somova"]
# __license__ = "MIT"
# __version__ = "0.1"
# __maintainer__ = "Daniele Antonucci"

# Libraries
import numpy as np
import pandas as pd
from sklearn.metrics import r2_score
from sklearn.linear_model import LinearRegression
# from pybuildingenergy.global_inputs import main_directory_
from pybuildingenergy.global_inputs import main_directory_
import pickle
import plotly.graph_objects as go


# ===== helper: restituisce (weekday, weekend) di 24 valori dal BUI o dai DEFAULT =====
def _get_pair_from_bui_or_default(bui_pair_or_none, default_weekday_24, default_weekend_24, name):
    """Ritorna due np.array(24): (weekday, weekend). Valida lunghezze e converte in float."""
    if bui_pair_or_none is not None:
        wd = np.asarray(bui_pair_or_none.get("weekday"), dtype=float)
        hd = np.asarray(bui_pair_or_none.get("weekend"), dtype=float)
    else:
        wd = np.asarray(default_weekday_24, dtype=float)
        hd = np.asarray(default_weekend_24, dtype=float)
    if wd.shape != (24,) or hd.shape != (24,):
        raise ValueError(f"{name}: profili devono essere 24 valori (weekday={wd.shape}, weekend={hd.shape})")
    return wd, hd

def _get_schedule_pair_for_bt(bt, workdays_map, weekend_map, name):
    """Estrae (weekday, weekend) dai dizionari di default in base al building_type_class (bt)."""
    if bt not in workdays_map or bt not in weekend_map:
        raise KeyError(f"{name}: '{bt}' non presente nei profili di default.")
    wd = np.asarray(workdays_map[bt], dtype=float)
    hd = np.asarray(weekend_map[bt], dtype=float)
    if wd.shape != (24,) or hd.shape != (24,):
        raise ValueError(f"{name}: profili default devono avere 24 valori.")
    return wd, hd

def _nan_to_zero(x):
    x = np.asarray(x, dtype=float)
    x[~np.isfinite(x)] = 0.0
    return x


# ===========================================================================================
def is_float(value):
    """
    Check if a value is float
    """
    try:
        float_value = float(value)
        return isinstance(float_value, float)
    except ValueError:
        return False


def capitalize_first_letter(s):
    """
    Capitalize first letter of string
    """
    if not s:
        return s
    return s[0].upper() + s[1:]


# ===========================================================================================
def Equation_of_time(day_of_year: pd.Series) -> pd.Series:
    """
    Equation of time according to formula 6.4.1.2 of ISO52010

    :param day_of_year: series of days in a year
    :return **t_eq**: equation of time
    """
    t_eq = 0.45 * (day_of_year - 359)
    mask = day_of_year < 21
    t_eq[mask] = 2.6 + 0.44 * day_of_year[mask]
    mask = (day_of_year >= 21) & (day_of_year < 136)
    t_eq[mask] = 5.2 + 9 * np.cos((day_of_year[mask] - 43) * 0.0357)
    mask = (day_of_year >= 136) & (day_of_year < 241)
    t_eq[mask] = 1.4 - 5 * np.cos((day_of_year[mask] - 135) * 0.0449)
    mask = (day_of_year >= 241) & (day_of_year < 336)
    t_eq[mask] = -6.3 - 10 * np.cos((day_of_year[mask] - 306) * 0.036)

    return t_eq


def Hour_angle_calc(solar_time: pd.Series) -> pd.Series:
    """
    Equation of time according to formula 6.4.1.5 of ISO52010

    :param solar_time: defined by formula 6.4.1.4 of ISO52010

    .. note::
        South is 0, East pos; (because of shading sectors: keep between -180 and + 180)

    :return **hour_angle_deg** hourly angle degrees
    
    """
    hour_angle_deg = 180 / 12 * (12.5 - solar_time)
    hour_angle_deg[hour_angle_deg > 180] -= 360
    hour_angle_deg[hour_angle_deg < -180] += 360

    return hour_angle_deg


def Air_mass_calc(solar_altitude_angle: pd.Series) -> pd.Series:
    """
    Calculation of air mass

    :param solar_altitude_angle: the angular distance between the rays of Sun and the horizon of the Earth

    :return **air_mass**

    """
    solar_altitude_angle_deg = np.degrees(solar_altitude_angle)
    air_mass = 1 / (
        np.sin(solar_altitude_angle)
        + 0.15 * np.power(solar_altitude_angle_deg + 3.885, -1.253)
    )
    mask = solar_altitude_angle_deg >= 10
    air_mass[mask] = 1 / np.sin(solar_altitude_angle[mask])

    return air_mass


def Get_positions(lst, value):
    """
    Ger position of a value
    """
    positions = []
    for i, item in enumerate(lst):
        if item == value:
            positions.append(i)
    return positions


def Filter_list_by_indices(lst, indices):
    """
    Filter a list by indices
    """
    return [lst[i] for i in indices]


def Perimeter_from_area(area, side):
    """
    Perimeter from area assuming 10m of side

    :param area: gross surface area of the slab on ground floor in [m]
    :param side: side of a rectangular shape of the building
    """
    base = area / side
    perimeter = 2 * (base + side)
    return perimeter


def Area_roof(leng_small, leng_roof):
    """
    Calculation of the roof area. According to the following formula:

    formula = ((length_small*cos(28)+0.5*cos(28))*(length_roof+(0.5*2)))*2

    **where**:

    cos(28°) = 0.88

    .. note:: 
        The calculation assumes that the building has a rectangular floor plan.

    :param leng_small: small length of roof side in [m]
    :param leng_roof: length of roof side in [m]

    """
    Area = 2 * ((leng_small * 0.88) + (0.5 * 0.88)) * (leng_roof + (0.5 * 2))
    return Area


def Internal_gains(bui_type: str, area: float):
    """
    Calculation of internal gains according to the building typology

    :param bui_type: type of building. Possible choice: residential, office, commercial
    :param area: gross area of the building

    :return: int_gains: value of internal gains in W/m2

    .. note::
        Only for residential building the data is avaialble.
        The result is a sum of contribution given by equipments, people heat and lights
    """
    if bui_type == "residential":
        int_gains = (120 + 100 + 2000 + 800 + 5 + 4 * 100) / area + 5 * 5

    else:
        int_gains = 5
    return int_gains


def Power_heating_system(bui_volume, bui_class):
    """
    Approssimative calculation of generator power, according to the following formula:
    p = Voume[m3] * energy needs[kW/m3]

    :param bui_class: could be:
        * *'old'*: No or very low insulated building
        * *'new'*: very well insulated building
        * *'average'*:medium insulated building

    :return heat_power: power of generator [W]
    """
    if bui_class == "old":
        p = bui_volume * 0.12
    elif bui_class == "gold":
        p = bui_volume * 0.03
    else:
        p = bui_volume * 0.05

    return p * 1000


def Check_area(elements_area: list):
    """
    Check the areas of the opaque and transparent components for the same orientation.
    If the opaque component is smaller than the transparent one, then the two areas are equalized.
    If the area of the floor slab on ground is higher than the area of the roof, that the two areas are equalized.

    The check follows the order of entry of the component as per the following reference.
    ["OP", "OP", "OP", "OP", "GR", "OP", "W", "W", "W", "W"]
    ['NV', 'SV', 'EV', 'WV', 'HOR', 'HOR', 'NV', 'SV', 'EV', 'WV']

    **where**:

    * OP: Opaque elements
    * GR: Ground Floor
    * W: Transaprent elements
    * NV: Vertical North
    * SV: South Vertical
    * EV: East Vertical
    * WV: West Vertical
    * HOR: Horizontal for ground floor(HOR-GR), Horizontal or Slope for Roof(HOR-OP)

    :param elements_area: List of areas for each individual facade element.

    :return check_el_area: list of corrected areas
    """
    check_el_area = elements_area.copy()
    for i in range(4):
        area_wall = elements_area[i]
        area_window = elements_area[i + 6]
        if area_wall < area_window:
            check_el_area[i] = area_window

    area_floor = check_el_area[4]
    area_roof = check_el_area[5]
    if area_roof < area_floor:
        check_el_area[4] = area_roof

    return check_el_area


#
# df_101_heating_hdd = df_101.loc[:,['energy_consumption_m3','heating_degree_days']]
# data_101 = [list(tuple(x)) for x in df_101_heating_hdd.itertuples(index=False, name=None)]


def Heating_Degree_days(out_temp: list, base_temperature: int) -> list:
    """
    Getting heating degree days from daily outdoor temperature


    :param base_temperature: outoddor base temperature to be used as limit for calculation of Heating Degree days.
            e.g.  18  in Europe
            e.g   20  in Italy
    :return HDD: Heating Degree days
    """

    HDD = []
    for temp in out_temp:
        if temp < base_temperature:
            HDD_value = base_temperature - temp
        else:
            HDD_value = 0

        HDD.append(HDD_value)
    return HDD


def check_dict_format(**kwargs):
    for arg_name, arg_value in kwargs.items():
        if not isinstance(arg_value, dict):
            raise TypeError(f"Argument '{arg_name}' must be a dictionary.")
    # If all inputs are in dictionary format, return True or do other operations
    return True


# ==========
def ePlus_shape_data(ep_result: pd.DataFrame, A_use: float):
    """
    Reshape dataset from energyplus:

    :param ep_result: dataframe of energy plus result
    :param A_use: net useful area in m2

    :return
        * **EnergyPlus_monthly_heating_in_kWh_per_sqm**: Monthly heating consumption of the building in kWh per m2
        * **EnergyPlus_monthly_cooling_in_kWh_per_sqm**: Monthly cooling consumption of the building in kWh per m2
        * **EnergyPlus_annual_heating_in_kWh_per_sqm**: Annual heating consumption of the building in kWh per m2
        * **EnergyPlus_annual_cooling_in_kWh_per_sqm**: Annual cooling consumption of the building in kWh per m2
    """

    ep_hourly_heating_in_J = ep_result[
        "ZONE ONE:Zone Air System Sensible Heating Energy [J](Hourly)"
    ]
    ep_hourly_cooling_in_J = ep_result[
        "ZONE ONE:Zone Air System Sensible Cooling Energy [J](Hourly) "
    ]  # Note: for some reason, EnergyPlus saves last output in csv with an extra space :-(
    # ep_hourly_T_op = ep_result['ZONE ONE:Zone Operative Temperature [C](Hourly)']

    month_start_hours = [
        0,
        744,
        1488,
        2160,
        2904,
        3624,
        4368,
        5088,
        5832,
        6576,
        7296,
        8040,
        8760,
    ]

    ep_monthly_heating_in_J = np.array(
        [
            ep_hourly_heating_in_J[
                month_start_hours[i] : month_start_hours[i + 1]
            ].sum()
            for i in range(12)
        ]
    )
    ep_monthly_cooling_in_J = np.array(
        [
            ep_hourly_cooling_in_J[
                month_start_hours[i] : month_start_hours[i + 1]
            ].sum()
            for i in range(12)
        ]
    )

    ep_annual_heating_in_J = ep_hourly_heating_in_J.sum()
    ep_annual_cooling_in_J = ep_hourly_cooling_in_J.sum()

    ep_annual_heating_in_kWh = ep_annual_heating_in_J / 3.6e6
    ep_annual_cooling_in_kWh = ep_annual_cooling_in_J / 3.6e6

    ep_monthly_heating_in_kWh = ep_monthly_heating_in_J / 3.6e6
    ep_monthly_cooling_in_kWh = ep_monthly_cooling_in_J / 3.6e6

    # ep_monthly_T_op = np.array([ep_hourly_T_op[month_start_hours[i]:month_start_hours[i + 1]].mean() for i in range(12)])

    EnergyPlus_annual_heating_in_kWh_per_sqm = ep_annual_heating_in_kWh / A_use
    EnergyPlus_annual_cooling_in_kWh_per_sqm = ep_annual_cooling_in_kWh / A_use

    EnergyPlus_monthly_heating_in_kWh_per_sqm = ep_monthly_heating_in_kWh / A_use
    EnergyPlus_monthly_cooling_in_kWh_per_sqm = ep_monthly_cooling_in_kWh / A_use

    return (
        EnergyPlus_monthly_heating_in_kWh_per_sqm,
        EnergyPlus_monthly_cooling_in_kWh_per_sqm,
        EnergyPlus_annual_heating_in_kWh_per_sqm,
        EnergyPlus_annual_cooling_in_kWh_per_sqm,
    )


def get_buildings_demos():
    """
    Get archetypes and demo buildings
    """
    # pickle_file_path = main_directory_ + "/pybuildingenergy/pybuildingenergy/data/archetypes.pickle"
    pickle_file_path = main_directory_ + "/archetypes.pickle"
    with open(pickle_file_path, "rb") as f:
        archetypes = pickle.load(f)

    return archetypes


def Simple_regeression(x_data: list, y_data: list, x_data_name: str):
    """
    Simple regression between two variable

    :param x_data: data on x-axes
    :param y_data: data on y-axes

    :return
        * **R2**: coefficient of determination
        * **equation**: linear regression equation
        * **residual**: residual !!! *Not yet included* !!!
    
    """
    #
    # Sample data
    x = np.array(x_data).reshape(-1, 1)  # Example independent variable
    y = np.array(y_data)  # Example dependent variable

    # Fit linear regression model
    model = LinearRegression().fit(x, y)

    # Predict y values using the fitted model
    y_pred = model.predict(x)

    # Calculate R-squared
    r2 = r2_score(y, y_pred)

    # Get the slope (coefficient) and intercept of the regression line
    slope = model.coef_[0]
    intercept = model.intercept_

    # Print R-squared and equation of the regression line
    # print("R-squared:", r2)
    # print("Equation of the regression line: y =", slope, "* x +", intercept)
    equation = f"y ={round(intercept,2)} + {x_data_name}*{round(slope,2)}"

    return (round(r2, 2), equation)


import numpy as np

def shading_reduction_factor(
    alpha_sol_t,           # [deg] solar altitude angle at time t
    phi_sol_t,             # [deg] solar azimuth angle at time t
    beta_k_t,              # [deg] window tilt (90 = vertical)
    gamma_k_t,             # [deg] window azimuth
    D_k_ovh_q,             # [m] horizontal overhang depth
    L_k_ovh_q,             # [m] setback/clear distance (or vertical element)
    elements_shading_type, # str | list[str] | None – types of shading elements
    H_k,                   # [m] window height
    W_k                    # [m] window width
):
    """
    Calculation of shading reduction factor
    :param alpha_sol_t: solar altitude angle at time t
    :param phi_sol_t: solar azimuth angle at time t
    :param beta_k_t: window tilt (90 = vertical)
    :param gamma_k_t: window azimuth
    :param D_k_ovh_q: horizontal overhang depth
    :param L_k_ovh_q: setback/clear distance (or vertical element)
    :param elements_shading_type: types of shading elements
    :param H_k: window height
    :param W_k: window width

    :return: (F_sh_dir_k_ts, h_k_sun_t)
      * **F_sh_dir_k_ts** in [0..1]: direct-beam shading reduction factor for the window at time t
      * **h_k_sun_t** [m]: remaining 'sunlit' vertical height on the window at time t
    """

    # ---- helper for safe numeric conversion ----
    def to_float(x, default=None):
        try:
            return float(x)
        except (TypeError, ValueError):
            return default

    # ---- minimal validation for window dimensions ----
    H = to_float(H_k, default=None)
    W = to_float(W_k, default=None)
    if H is None or W is None or H <= 0 or W <= 0:
        # without valid window dimensions the calculation is not meaningful
        return 0.0, 0.0

    # ---- angle checks in degrees (as in your original logic) ----
    a = to_float(alpha_sol_t, default=None)
    p = to_float(phi_sol_t,   default=None)
    b = to_float(beta_k_t,    default=None)
    g = to_float(gamma_k_t,   default=None)
    if None in (a, p, b, g):
        return 0.0, 0.0

    # if sun is far from the window normal/tilt, no direct sun on the pane
    if abs(g - p) > 90 or abs(b - a) > 90:
        return 0.0, 0.0

    # ---- convert to radians for trigonometric functions ----
    alpha = np.radians(a)
    phi   = np.radians(p)
    gamma = np.radians(g)

    # ---- normalize shading types: str/list -> lowercase set ----
    types = set()
    if elements_shading_type:
        if isinstance(elements_shading_type, str):
            types = {elements_shading_type.lower()}
        else:
            try:
                types = {str(t).lower() for t in elements_shading_type}
            except TypeError:
                types = {str(elements_shading_type).lower()}

    has_overhang  = any(t in types for t in ("horizontal_overhang", "vertical_overhang"))
    has_obstacles = "obstacles" in types

    h_k_sun_t = 0.0

    # ---- overhangs (horizontal/vertical projections) ----
    if has_overhang:
        D = to_float(D_k_ovh_q, default=0.0)
        L = to_float(L_k_ovh_q, default=0.0)

        # projection denominator; protect against ~0 to avoid division by zero
        denom = np.cos(phi - gamma)
        if abs(denom) < 1e-12:
            denom = 1e-12 * np.sign(denom if denom != 0 else 1.0)

        # geometric projection of the overhang shadow height on the window
        h_k_ovh_q_t = D * np.tan(alpha) / denom - L
        h_k_ovh_q_t = max(0.0, h_k_ovh_q_t)

        # remaining sunlit height after overhang shadow
        h_k_sun_t   = max(0.0, H - h_k_ovh_q_t)

    # ---- external obstacles (if you have a model, replace the placeholder) ----
    if has_obstacles:
        h_k_obst_p_t = 0.0  # TODO: replace with your obstacle-shadow calculation/parameter
        h_k_sun_t = max(0.0, H - h_k_obst_p_t)

    # if no shading element affected the height, assume fully sunlit height
    if not (has_overhang or has_obstacles):
        h_k_sun_t = H

    # sunlit width: if you do not model horizontal shading extent, use total width
    w_k_sun_t = W

    # shading reduction factor based on sunlit area over total area
    F_sh_dir_k_ts = (h_k_sun_t * w_k_sun_t) / (H * W)
    F_sh_dir_k_ts = float(np.clip(F_sh_dir_k_ts, 0.0, 1.0))

    return F_sh_dir_k_ts, h_k_sun_t


# ====================================================================================================
#                                       DHW Table of ISO 12831
# ====================================================================================================

# Hourly breakdown of relative demand for hot water by volume. Table B.1
table_B_1 = pd.DataFrame(
    {
        "XXS" : [0, 0, 0,0, 0, 0, 0, 10, 5, 5, 0, 10, 15, 0, 0, 0, 0, 0, 15, 10, 10, 20, 0, 0],
        "XS" : [0, 0, 0, 0, 0, 0, 0, 25, 0, 0, 0, 0, 25, 0, 0, 0, 0, 0, 0, 0, 50, 0, 0, 0],
        "S" : [0, 0, 0, 0, 0, 0, 0, 10, 5, 5, 0, 10, 15, 0, 0, 0, 0, 0, 10, 0, 20, 25, 0, 0],
        "M": [0, 0, 0, 0, 0, 0, 0, 27.5, 7.2, 3.6, 1.8, 3.6, 5.4, 0, 1.8, 1.8, 1.8, 0, 5.4, 1.8, 12.6, 25.7, 0, 0],
        "L": [0, 0, 0, 0, 0, 0, 0, 14.7, 33.6, 1.8, 0.9, 1.8, 2.7, 0, 0.9, 0.9, 0.9, 0, 2.7, 0.9, 6.3, 31.8, 0, 0],
        "XL": [0, 0, 0, 0, 0, 0, 0, 33.8, 2.2, 1.1, 1.1, 1.7, 3.9, 0, 0.6, 1.1, 1.1, 0.6, 1.7, 0.6, 27, 23.7, 0, 0 ],
        "XXL": [0, 0, 0, 0, 0, 0, 0, 33.7, 1.7, 0.9, 0.9, 1.3, 3, 0, 0.4, 0.9, 0.9, 0.4, 1.3, 0.4, 28.4, 25.9, 0, 0]
     },
     index = [list(range(1,25))]
)

#Hourly distribution of relative volume-based hot water demand across various building categories. Table B.2
table_B_2 = pd.DataFrame(
    {
        "single_family_dwelling" : [1.8, 1, 0.6, 0.3, 0.4, 0.6, 2.4, 4.7, 6.8, 5.7, 6.1, 6.1, 6.3, 6.4, 5.1, 4.4, 4.3, 4.7, 5.7, 6.5, 6.6, 5.8, 4.5, 3.1],
        "appartment dwelling" : [1, 1, 1, 0, 0, 1, 3, 6, 8, 6, 5, 5, 6, 6, 5, 4, 4, 5, 6, 7, 7, 6, 5, 2],
        "residential_home_for_the_elderly" : [0.3, 0.3, 0.4, 0.7, 1.0, 1.8, 9.3, 15.7, 8.1, 7.5, 7.0, 6.6, 7.1, 5.1, 3.8, 3.3, 4.1, 2.9, 6.1, 4.1, 1.4, 1.8, 0.9, 0.4],
        "student_residence": [1.4, 1.0, 0.5, 0.6, 1.3, 3.4, 5.8, 5.8, 6.2, 5.4, 5.1, 4.7, 4.2, 4.5, 4.1, 4.3, 5.3, 6.0, 6.6, 6.0, 5.6, 5.4, 3.9, 2.8],
        "Hospital": [0.4, 0.4, 0.5, 0.8, 1.2, 2.8, 7.5, 10.5, 8.0, 7.5, 7.5, 7.0, 7.5, 5.5, 4.3, 3.7, 4.5, 3.2, 7.0, 4.5, 2.0, 2.0, 1.2, 0.5]
     },
     index = [list(range(1,25))]
)

#  Net energy demand for water heating per day. Table B.3  
table_B_3 = pd.DataFrame({
    'type_of_usage' : ['Office_buildings', "Hospital ward or patient's room", 
                       'School_without_showers', 'School_with_shower',
                       'Retail shop/department store', 'Workshop, industrial facility (for washing snd showering)',
                       'Modest hotel', 'Medium-class hotel',
                       'Luxury-class hotel', 'Restaurant, inn/pub',
                       'Home(for the aged, orphanage, etc.)', 'Barracks',
                       'Sport faciltiy with showers', 'Commercial catering kitchen',
                       'Bakery', 'Hairdresser/barber',
                       'Butcher with production',
                       'Laundry', 'Brewery', 'Dairy'],
    'Usage' : ['person', 'bed', 'person', 'person', 'employee', 'employee', 
               'bed', 'bed', 'bed', 'seat', 'person', 'person', 'person', 'meal',
               'employee', 'employee', 'employee', '100 Kg laundry', '100 liters beer', 
               '100 liters of milk'],
    'Usage dependent' : [0.4, 8, 0.5, 1.5, 1, 1.5, 1.5, 4.5, 7, 1.5, 3.5, 1.5, 1.5, 0.4, 5, 8, 18, 20, 15, 10],
    'Area specific - Wh/m2d': [30, 530, 170, 500, 10, 75, 190, 450, 580, 1250, 230, 150, None, None, None, None, None, None, None, None],
    'Reference Area': ['office floor area', 'wards and room', 'classrooms', 'classrooms',
                       'sales areas', 'area of workshop/works', 'hotel bedrooms',
                       'hotel bedrooms', 'hotel bedrooms', 'public rooms', 'rooms', 'rooms', 
                       None, None, None, None, None, None, None, None], 

})

# Values for the calculation of domestic hot watrer requirements per day
table_B_4 = pd.DataFrame(
    {
    'type_of_activity'  : [
        'Accomodation', 'Health establishment wihtout accomodation', 
        'Health establishment without accomodation',
        'Health establishment without accomodation - without laundry',
        'Catering, 2 meals per day. Traditional cusine',
        'Catering, 2 meals per day. Self service', 
        'Catering, 1 meals per day. Tradional cusine', 
        'Catering, 1 meals per day. Self service',
        'Hotel, 1-star without laundry', 'Hotel, 1-star withlaundry', 
        'Hotel, 2-star without laundry', 'Hotel, 2-star withlaundry', 
        'Hotel, 3-star without laundry', 'Hotel, 3-star withlaundry', 
        'Hotel, 4-star and GC without laundry', 'Hotel, 4-star and GC withlaundry', 
        'Sport_establishment'
        ],
    'V_W_f_day' : [28, 10, 56, 88, 21, 8, 10, 4, 56, 70, 76, 90, 97, 111, 118, 132, 101 ], # [l/d]
    'f' : ['number of beds'] * 4 + ['number of guest per meal'] * 4 + ['number of beds'] * 8 + ['number_of_showers_installed']
    }
)

# values for calculation of domestic hot water requirements per day
table_B_5_Standard = pd.DataFrame({
    'type_pf_building': ['residential_buildings (simple_housing)', 'residential_buildings (luxury_building)',
                         'single_family_dwellings', 'apartment_dwellings'],
    'V_W_p_day':['25-60','60-100', '40-70', '25-30']
})

# values for calculation of domestic hot water requirements per day as avergae of the values from table_B_5_standard
table_B_5_modified = pd.DataFrame({
    'type_of_building': [
        'residential_building - simple housing - MIN', 
        'residential_building - simple housing - AVG', 
        'residential_building - simple housing - MAX', 
        'residential_building - luxury housing - MIN', 
        'residential_building - luxury housing - AVG', 
        'residential_building - luxury housing - MAX', 
        'single_family_dwellings - MIN', 
        'single_family_dwellings - AVG', 
        'single_family_dwellings - MAX', 
        'apartments_dwellings - MIN', 
        'apartments_dwellings - AVG', 
        'apartments_dwellings - MAX', 
        ],
    'liters/person_per_day':[25, 45, 60, 60, 80, 100, 40, 55, 70, 25, 27.5, 30]
})


def plot_sankey_building(sankey_data):
    '''
    Plot sankey graph of the energy balance

    :param sankey_data:data set of energy balance
    :return: sankey graph

    '''
    inputs = sankey_data.get("inputs", {})
    outputs = sankey_data.get("outputs", {})
    storage = sankey_data.get("energy_accumulated_zone", 0.0)

    def _finite_pos(x): 
        try:
            x = float(x)
            return np.isfinite(x) and abs(x) > 0.0
        except Exception:
            return False

    in_labels  = [k for k, v in inputs.items()  if _finite_pos(v)]
    in_values  = [float(inputs[k])  for k in in_labels]
    out_labels = [k for k, v in outputs.items() if _finite_pos(v)]
    out_values = [float(outputs[k]) for k in out_labels]

    node_labels = in_labels + ["Zone"] + out_labels
    node_index  = {lab: i for i, lab in enumerate(node_labels)}

    sources = [node_index[lab] for lab in in_labels]
    targets = [node_index["Zone"]] * len(in_labels)
    values  = in_values[:]

    # Add storage flow if present
    if _finite_pos(storage):
        node_labels.append("Storage")
        node_index["Storage"] = len(node_labels) - 1
        if storage > 0:
            # energy stored (Zone → Storage)
            sources += [node_index["Zone"]]
            targets += [node_index["Storage"]]
            values  += [float(storage)]
        else:
            # energy released from storage (Storage → Zone)
            sources += [node_index["Storage"]]
            targets += [node_index["Zone"]]
            values  += [abs(float(storage))]

    sources += [node_index["Zone"]] * len(out_labels)
    targets += [node_index[lab] for lab in out_labels]
    values  += out_values

    fig = go.Figure(data=[go.Sankey(
        arrangement="snap",
        node=dict(label=node_labels, pad=15, thickness=18),
        link=dict(source=sources, target=targets, value=values)
    )])
    fig.update_layout(title="Annual energy balance — Sankey", font_size=12)
    return fig



from copy import deepcopy
import numpy as np
from pybuildingenergy.global_inputs import TB14 as TB14_backup

def _dir_from_orientation(azimuth: float, tilt: float) -> str:
    '''
    Map an orientation to a cardianl code used for adjacency logic

    Args:
        azimuth (float): azimuth angle of the surface
        tilt (float): tilt angle of the surface

    Returns:
        str: one of "HOR"(horizontal), "NV"(north vertical), "EV"(east vertical), "SV"(south vertical), "WV"(west vertical)
    '''
    # 0=N, 90=E, 180=S, 270=W; tilt 90 = verticale; 0 = orizzontale
    if abs(float(tilt) - 0.0) <= 1e-9:
        return "HOR"
    if abs(float(tilt) - 90.0) > 1e-6:
        # fallback: soglia 45°
        return "HOR" if float(tilt) < 45.0 else "NV"
    az = float(azimuth) % 360.0
    if abs(az-0.0) <= 1e-6 or abs(az-360.0) <= 1e-6: return "NV"
    if abs(az-90.0) <= 1e-6:  return "EV"
    if abs(az-180.0) <= 1e-6: return "SV"
    if abs(az-270.0) <= 1e-6: return "WV"
    # snap al più vicino
    pts  = [0.0, 90.0, 180.0, 270.0]
    labs = ["NV","EV","SV","WV"]
    return labs[int(np.argmin(np.abs((az - np.array(pts)) % 360.0)))]


def _adj_op_area_in_dir(adj_zone: dict, dir_code: str) -> float:
    A   = np.asarray(adj_zone["area_facade_elements"], dtype=float)
    Typ = np.asarray(adj_zone["typology_elements"], dtype=object)
    Ori = np.asarray(adj_zone["orientation_elements"], dtype=object)
    mask = (np.char.upper(Typ.astype(str)) == "OP") & (np.char.upper(Ori.astype(str)) == dir_code)
    return float(A[mask].sum()) if mask.any() else 0.0

#  -------------------------------------------------------------------
#  CHECK INPUT HEATING SYSTEM
# -------------------------------------------------------------------

import pandas as pd

def check_heating_system_inputs(system_input: dict):
    """
    Upload and validate the heating system configuration from a dict.

    Rules:
    - If 'TB14' is present and DataFrame -> use that.
    - Otherwise -> use the backup TB14 from global_inputs.py.
    - If emitter_type is not an index of the TB14 selected:
        -> set automatico al primo indice disponibile e messaggio di avviso.
    - If gen_flow_temp_control_type = Type A, then gen_outdoor_temp_data must be provided as a dataframe.
    - If gen_flow_temp_control_type = Type C, if θHW_gen_flw_const is not provided, it is set to 50.

    Returns:
        {
            "TB14_used": DataFrame,
            "emitter_type": str,
            "messages": list[str],
            "config": dict  # input normalizzato con emitter_type aggiornato
        }
    """
    messages = []
    cfg = deepcopy(system_input)

    # 1) TB14: custom o default
    tb14_custom = cfg.get("TB14", None)
    if isinstance(tb14_custom, pd.DataFrame):
        TB14 = tb14_custom
        messages.append("✅ Custom TB14 table loaded from input.")
    else:
        TB14 = TB14_backup
        messages.append("⚙️ Default TB14 table loaded from global_inputs.py.")

    # 2) Validazione emitter_type rispetto alla TB14 usata
    emitter_type = cfg.get("emitter_type", None)
    valid_emitters = list(TB14.index)

    if emitter_type not in valid_emitters:
        fallback = valid_emitters[0] if valid_emitters else None
        messages.append(
            f"⚠️ Emitter type '{emitter_type}' not found in TB14; auto-set to '{fallback}'."
        )
        cfg["emitter_type"] = fallback
    else:
        messages.append(f"✅ Emitter type '{emitter_type}' found in TB14.")

    # 3) Validazione gen_flow_temp_control_type (Type A, B, or C)
    gen_flow_temp_control_type = cfg.get("gen_flow_temp_control_type", None)
    
    if gen_flow_temp_control_type == 'Type A':
        gen_outdoor_temp_data = cfg.get('gen_outdoor_temp_data', None)
        if not isinstance(gen_outdoor_temp_data, pd.DataFrame):
            messages.append("⚠️ For 'Type A' control, 'gen_outdoor_temp_data' must be provided as a DataFrame.")
            cfg["gen_flow_temp_control_type"] = 'Type B'  # Auto-switch to Type B
        else:
            messages.append("✅ 'gen_outdoor_temp_data' provided for 'Type A' control.")
    
    elif gen_flow_temp_control_type == 'Type C':
        if 'θHW_gen_flw_const' not in cfg or cfg['θHW_gen_flw_const'] is None:
            messages.append("⚠️ 'θHW_gen_flw_const' not provided for 'Type C' control; setting it to 50.")
            cfg['θHW_gen_flw_const'] = 50.0
        else:
            messages.append(f"✅ 'θHW_gen_flw_const' provided: {cfg['θHW_gen_flw_const']} for 'Type C' control.")
    
    elif gen_flow_temp_control_type != 'Type B':
        messages.append(f"⚠️ Invalid value for 'gen_flow_temp_control_type': '{gen_flow_temp_control_type}'. Setting to 'Type B'.")
        cfg["gen_flow_temp_control_type"] = 'Type B'

    return {
        "TB14_used": TB14,
        "emitter_type": cfg["emitter_type"],
        "messages": messages,
        "config": cfg
    }




def sanitize_and_validate_BUI(bui: dict,
                              fix: bool = True,
                              eps: float = 1e-6,
                              defaults: dict | None = None):
    """
    Validates and (optionally) sanitizes a BUI, with extra checks:
    - Check area OP (NV/EV/SV/WV) of the adjacent zone with respect to the BUI wall 'adjacent';
    - Check windows in the same direction of the 'adjacent' wall (windows cannot exceed
      the difference between the area of the BUI 'adjacent' and the area of the OP of the adjacent zone);
    - If the adjacent zone covers less area than the BUI 'adjacent', creates a new
      'opaque' element with area equal to the difference and resizes the 'adjacent' wall to the sole area covered.
    """

    bui_clean = deepcopy(bui)
    issues = []
    D = defaults or {
        "opaque_u_default": 0.5,
        "transparent_u_default": 1.6,
        "g_default": 0.6
    }

    def add_issue(level, path, msg, fixed=False):
        issues.append({"level": level, "path": path, "msg": msg, "fix_applied": fixed})

    # ---------------------------
    # 1) BUILDING-LEVEL CHECKS
    # ---------------------------
    b = bui_clean.get("building", {})
    for key in ["net_floor_area", "exposed_perimeter", "height"]:
        if key in b:
            val = b[key]
            if not (isinstance(val, (int, float)) and val > 0):
                if fix and isinstance(val, (int, float)) and val == 0:
                    b[key] = max(eps, 1.0)
                    add_issue("WARN", f"building.{key}", f"{key} era 0; impostato a {b[key]}", fixed=True)
                else:
                    add_issue("ERROR", f"building.{key}", f"{key} deve essere > 0; valore={val}", fixed=False)

    if "n_floors" in b and (not isinstance(b["n_floors"], (int, float)) or b["n_floors"] <= 0):
        if fix and isinstance(b["n_floors"], (int, float)) and b["n_floors"] == 0:
            b["n_floors"] = 1
            add_issue("WARN", "building.n_floors", "n_floors era 0; impostato a 1", fixed=True)
        else:
            add_issue("ERROR", "building.n_floors", f"n_floors deve essere > 0; valore={b.get('n_floors')}", fixed=False)

    # ---------------------------
    # 2) SURFACES CHECKS
    # ---------------------------
    type_corrections = {"opque": "opaque", "opaqu": "opaque", "trasparent": "transparent"}
    allowed_types = {"opaque", "transparent", "adiabatic", "adjacent"}

    for i, s in enumerate(bui_clean.get("building_surface", [])):
        path = f"building_surface[{i}]"
        # correzione refusi type
        t = s.get("type")
        if isinstance(t, str) and t.lower() in type_corrections:
            old = t
            s["type"] = type_corrections[t.lower()]
            add_issue("WARN", f"{path}.type", f'type="{old}" corretto in "{s["type"]}"', fixed=True)
        t = (s.get("type") or "").lower()
        if t not in allowed_types:
            add_issue("ERROR", f"{path}.type", f'type "{s.get("type")}" non riconosciuto', fixed=False)

        # area
        area = s.get("area", None)
        if not (isinstance(area, (int, float)) and area > 0):
            if fix and isinstance(area, (int, float)) and area == 0:
                s["area"] = max(eps, 1.0)
                add_issue("WARN", f"{path}.area", f"area era 0; impostata a {s['area']}", fixed=True)
            else:
                add_issue("ERROR", f"{path}.area", f"area deve essere > 0; valore={area}", fixed=False)

        # sky_view_factor
        svf = s.get("sky_view_factor", None)
        if not (isinstance(svf, (int, float)) and 0.0 <= svf <= 1.0):
            if fix and isinstance(svf, (int, float)):
                s["sky_view_factor"] = min(1.0, max(0.0, svf))
                add_issue("WARN", f"{path}.sky_view_factor", f"sky_view_factor fuori range; clippato a {s['sky_view_factor']}", fixed=True)
            else:
                add_issue("WARN", f"{path}.sky_view_factor", f"sky_view_factor mancante o non in [0,1]; valore={svf}", fixed=False)

        # U-value per opaque/transparent
        if t in {"opaque", "transparent"}:
            u = s.get("u_value", None)
            if not (isinstance(u, (int, float)) and u > 0):
                if fix:
                    default_u = D["opaque_u_default"] if t == "opaque" else D["transparent_u_default"]
                    s["u_value"] = default_u
                    add_issue("WARN", f"{path}.u_value", f"u_value non valido ({u}); impostato default {default_u}", fixed=True)
                else:
                    add_issue("ERROR", f"{path}.u_value", f"u_value deve essere > 0; valore={u}", fixed=False)

        # g-value per trasparenti
        if t == "transparent":
            g = s.get("g_value", None)
            if not (isinstance(g, (int, float)) and g > 0):
                if fix:
                    s["g_value"] = D["g_default"]
                    add_issue("WARN", f"{path}.g_value", f"g_value non valido ({g}); impostato default {D['g_default']}", fixed=True)
                else:
                    add_issue("ERROR", f"{path}.g_value", f"g_value deve essere > 0; valore={g}", fixed=False)

            # dimensioni finestra
            h = s.get("height", None)
            w = s.get("width", None)
            if not (isinstance(h, (int, float)) and h > 0):
                if fix:
                    s["height"] = 1.0
                    add_issue("WARN", f"{path}.height", f"height non valido ({h}); impostato a 1.0", fixed=True)
                else:
                    add_issue("ERROR", f"{path}.height", f"height deve essere > 0; valore={h}", fixed=False)
            if not (isinstance(w, (int, float)) and w > 0):
                if fix:
                    s["width"] = 1.0
                    add_issue("WARN", f"{path}.width", f"width non valido ({w}); impostato a 1.0", fixed=True)
                else:
                    add_issue("ERROR", f"{path}.width", f"width deve essere > 0; valore={w}", fixed=False)

            # parapet 0..height
            para = s.get("parapet", None)
            if para is not None and isinstance(para, (int, float)) and h is not None and isinstance(h, (int,float)):
                if not (0.0 <= para <= max(h, 0.0)):
                    if fix:
                        s["parapet"] = min(max(0.0, para if isinstance(para, (int,float)) else 0.0), h if isinstance(h,(int,float)) else 1.0)
                        add_issue("WARN", f"{path}.parapet", f"parapet fuori range; clippato a {s['parapet']}", fixed=True)
                    else:
                        add_issue("WARN", f"{path}.parapet", f"parapet fuori range (0..height); valore={para}, height={h}", fixed=False)

        # orientamenti consigliati
        ori = s.get("orientation", {})
        tilt = ori.get("tilt", None)
        az   = ori.get("azimuth", None)
        if not (isinstance(tilt, (int, float)) and tilt in (0, 90)):
            if fix and isinstance(tilt, (int, float)):
                s["orientation"]["tilt"] = 0 if abs(tilt - 0) < abs(tilt - 90) else 90
                add_issue("WARN", f"{path}.orientation.tilt", f"tilt non standard ({tilt}); normalizzato a {s['orientation']['tilt']}", fixed=True)
            else:
                add_issue("WARN", f"{path}.orientation.tilt", f"tilt consigliato 0 o 90; valore={tilt}", fixed=False)
        if not (isinstance(az, (int, float)) and az in (0, 90, 180, 270)):
            if fix and isinstance(az, (int, float)):
                snapped = int(round(az / 90.0) * 90) % 360
                if snapped == 360: snapped = 0
                s["orientation"]["azimuth"] = snapped
                add_issue("WARN", f"{path}.orientation.azimuth", f"azimuth non standard ({az}); normalizzato a {snapped}", fixed=True)
            else:
                add_issue("WARN", f"{path}.orientation.azimuth", f"azimuth consigliato in {{0,90,180,270}}; valore={az}", fixed=False)

    # ---------------------------
    # 3) ADJACENT ZONES CHECKS
    # ---------------------------
    for j, az in enumerate(bui_clean.get("adjacent_zones", [])):
        path = f"adjacent_zones[{j}]"
        vol = az.get("volume", None)
        if not (isinstance(vol, (int, float)) and vol > 0):
            if fix and isinstance(vol, (int, float)) and vol == 0:
                az["volume"] = 1.0
                add_issue("WARN", f"{path}.volume", "volume era 0; impostato a 1.0", fixed=True)
            else:
                add_issue("ERROR", f"{path}.volume", f"volume deve essere > 0; valore={vol}", fixed=False)

        # coerenza lunghezze
        arrs = {
            "area_facade_elements": az.get("area_facade_elements"),
            "typology_elements": az.get("typology_elements"),
            "transmittance_U_elements": az.get("transmittance_U_elements"),
            "orientation_elements": az.get("orientation_elements"),
        }
        lengths = [len(v) for v in arrs.values() if isinstance(v, (list, np.ndarray))]
        if lengths and len(set(lengths)) != 1:
            add_issue("ERROR", path, f"Lunghezze arrays non coerenti: {[ (k, None if v is None else len(v)) for k,v in arrs.items() ]}", fixed=False)

        # U elementi > 0
        U = arrs.get("transmittance_U_elements")
        if isinstance(U, (list, np.ndarray)):
            U = np.asarray(U, dtype=float)
            bad = np.where(U <= 0)[0].tolist()
            if bad:
                if fix:
                    U[U <= 0] = D["opaque_u_default"]
                    az["transmittance_U_elements"] = U
                    add_issue("WARN", f"{path}.transmittance_U_elements", f"U<=0 ai idx {bad}; impostati a {D['opaque_u_default']}", fixed=True)
                else:
                    add_issue("ERROR", f"{path}.transmittance_U_elements", f"U<=0 ai idx {bad}", fixed=False)
    
    # -------------------------------------------------------------------
    # 4) ADJACENCY PAIRING CAPS + FINESTRE + SPLIT ESTERNO
    # -------------------------------------------------------------------
    def _azimuth_to_dir(az: float) -> str | None:
        if az is None:
            return None
        azn = int(round(float(az))) % 360
        if azn in (0, 360): return "NV"
        if azn == 90:       return "EV"
        if azn == 180:      return "SV"
        if azn == 270:      return "WV"
        return None

    adj_by_name = { z.get("name"): z for z in bui_clean.get("adjacent_zones", []) }

    i = 0
    while i < len(bui_clean.get("building_surface", [])):
        s = bui_clean["building_surface"][i]
        path = f"building_surface[{i}]"

        if (s.get("type") != "adjacent") or (not s.get("name_adj_zone")):
            i += 1
            continue

        ori = s.get("orientation", {}) or {}
        tilt = ori.get("tilt", None)
        az   = ori.get("azimuth", None)

        # Consideriamo solo pareti verticali
        if not (isinstance(tilt, (int,float)) and abs(tilt-90) < 1e-6):
            i += 1
            continue

        dir_card = _azimuth_to_dir(az)
        if dir_card is None:
            i += 1
            continue

        A_bui_orig = float(s.get("area", 0.0))
        if A_bui_orig <= 0:
            i += 1
            continue

        adj_name = s["name_adj_zone"]
        if adj_name not in adj_by_name:
            add_issue("ERROR", path, f"name_adj_zone='{adj_name}' non trovata in adjacent_zones", fixed=False)
            i += 1
            continue

        # --- arrays zona adiacente (writable) ---
        azone = adj_by_name[adj_name]
        A_arr  = np.array(azone.get("area_facade_elements", []), dtype=float).copy()
        TypArr = np.array(azone.get("typology_elements", []),     dtype=object)
        OriArr = np.array(azone.get("orientation_elements", []),  dtype=object)

        if not (len(A_arr) == len(TypArr) == len(OriArr)):
            add_issue("ERROR", f"adjacent_zones[{adj_name}]", "Arrays non coerenti (len mismatch) per il cap per direzione", fixed=False)
            i += 1
            continue

        # === (1) CAP area OP della zona adiacente sulla direzione ===
        mask_op_dir = (np.char.upper(TypArr.astype(str)) == "OP") & (np.char.upper(OriArr.astype(str)) == dir_card)
        idx = np.where(mask_op_dir)[0]
        if idx.size == 0:
            A_adj_dir_cap = 0.0
        else:
            A_adj_dir_cap = float(A_arr[idx].sum())
            if A_adj_dir_cap > A_bui_orig + eps:
                msg = (f"Zona '{adj_name}' OP {dir_card}={A_adj_dir_cap:.3f} m² > parete BUI {A_bui_orig:.3f} m² "
                       f"({s.get('name','(senza nome)')})")
                if not fix:
                    add_issue("ERROR", path, msg + " (no fix)", fixed=False)
                else:
                    sf = A_bui_orig / max(A_adj_dir_cap, eps)
                    A_arr[idx] = A_arr[idx] * sf
                    azone["area_facade_elements"] = A_arr  # write-back
                    A_adj_dir_cap = float(A_arr[idx].sum())
                    add_issue("WARN", path, msg + f" → ridotto (scale {sf:.3f})", fixed=True)

        # === (2) WINDOW CHECK nella stessa direzione (usa A_bui_orig e A_adj_dir_cap) ===
        win_idx = []
        for k, ws in enumerate(bui_clean.get("building_surface", [])):
            if (ws.get("type") == "transparent"
                and isinstance(ws.get("orientation",{}).get("tilt", None), (int,float))
                and abs(ws["orientation"]["tilt"] - 90) < 1e-6):
                if _azimuth_to_dir(ws["orientation"].get("azimuth", None)) == dir_card:
                    win_idx.append(k)

        if win_idx:
            A_win_dir = float(sum(bui_clean["building_surface"][k].get("area", 0.0) for k in win_idx))
            diff = max(A_bui_orig - A_adj_dir_cap, 0.0)

            if A_win_dir > diff + eps:
                msg = (f"Finestre {dir_card}: area finestre={A_win_dir:.3f} m² > differenza "
                       f"(A_BUI_adjacent - A_adj_OP)={diff:.3f} m² su {s.get('name','(senza nome)')}")
                if not fix:
                    add_issue("ERROR", path, msg + " (no fix)", fixed=False)
                else:
                    # scala proporzionalmente tutte le finestre in questa direzione per farle sommare a 'diff'
                    sfw = (diff / max(A_win_dir, eps)) if diff > 0 else 0.0
                    for k in win_idx:
                        oldA = float(bui_clean["building_surface"][k].get("area", 0.0))
                        newA = oldA * sfw
                        bui_clean["building_surface"][k]["area"] = newA
                    add_issue("WARN", path, msg + f" → aree finestre ridotte (scale {sfw:.3f})", fixed=True)

        # === (3) SPLIT: se la zona adiacente copre meno di A_bui_orig, crea nuova parete 'opaque' per il residuo ===
        diff_ext = max(A_bui_orig - A_adj_dir_cap, 0.0)
        if diff_ext > eps:
            if not fix:
                add_issue("ERROR", path, (f"Parete 'adjacent' {s.get('name','(senza nome)')} non coperta interamente dalla "
                                          f"zona '{adj_name}' ({A_adj_dir_cap:.3f} < {A_bui_orig:.3f}). "
                                          f"Residuo {diff_ext:.3f} m² non allocato."), fixed=False)
            else:
                # crea nuovo elemento OPAQUE per il residuo
                new_surf = {
                    "name": f"{s.get('name','(adjacent)')} — residuo esterno",
                    "type": "opaque",
                    "area": float(diff_ext),
                    "sky_view_factor": s.get("sky_view_factor", 0.5),
                    "u_value": s.get("u_value", D["opaque_u_default"]),
                    "solar_absorptance": s.get("solar_absorptance", 0.6),
                    "thermal_capacity": s.get("thermal_capacity", 1000),
                    "orientation": {
                        "azimuth": s.get("orientation", {}).get("azimuth", 0),
                        "tilt": s.get("orientation", {}).get("tilt", 90),
                    },
                    "name_adj_zone": None,
                }
                bui_clean["building_surface"].append(new_surf)

                # ridimensiona la parete adjacent alla sola area coperta dalla zona adiacente
                s["area"] = float(A_adj_dir_cap)

                add_issue("WARN", path, (f"Creato elemento 'opaque' residuo di {diff_ext:.3f} m² e "
                                         f"ridimensionata parete 'adjacent' a {A_adj_dir_cap:.3f} m²."), fixed=True)

        # passo al prossimo elemento (attenzione: la lista può essere cresciuta)
        i += 1

        # ----------------------------------------------------------
        # 4) SPLIT PARETI ADJACENT: parte adiacente + parte esterna
        #    Crea nuova superficie "opaque" per la porzione esterna.
        #    (ESEGUIRE DOPO TUTTI I CHECK)
        # ----------------------------------------------------------
        new_surfaces = []   # colleziono qui e aggiungo con .extend a fine loop
        for i, s in enumerate(bui_clean.get("building_surface", [])):
            if (s.get("type") != "adjacent") or (not s.get("name_adj_zone")):
                continue

            # Direzione NV/EV/SV/WV/HOR della superficie BUI
            az = s.get("orientation", {}).get("azimuth", 0.0)
            tl = s.get("orientation", {}).get("tilt", 90.0)
            dir_code = _dir_from_orientation(az, tl)

            # Consideriamo SOLO le verticali per lo split con zona adiacente
            if dir_code not in ("NV","EV","SV","WV"):
                continue

            A_bui = float(s.get("area", 0.0))
            if A_bui <= 0:
                continue

            # Trova la zona adiacente corrispondente per nome
            adj_name = s.get("name_adj_zone")
            adj_zone = None
            for azj in bui_clean.get("adjacent_zones", []):
                if azj.get("name") == adj_name:
                    adj_zone = azj
                    break
            if adj_zone is None:
                # Nome non trovato: segnalo
                add_issue("WARN", f"building_surface[{i}]", f'name_adj_zone="{adj_name}" non trovato tra le adjacent_zones', fixed=False)
                continue

            # Area OP della adj-zone nella stessa direzione
            A_adj_dir = _adj_op_area_in_dir(adj_zone, dir_code)

            # CAP: l’adiacente non può eccedere la parete BUI
            if A_adj_dir > A_bui:
                add_issue("WARN",
                        f"building_surface[{i}]",
                        f"Area zona adiacente {adj_name} in {dir_code} ({A_adj_dir:.3f} m²) > area BUI ({A_bui:.3f} m²); cappata a {A_bui:.3f}",
                        fixed=True)
                A_adj_dir = A_bui

            # Se la zona adiacente è PIÙ PICCOLA della parete BUI → crea residuo esterno
            diff = A_bui - A_adj_dir
            if diff > 1e-9:
                # 1) riduci la superficie ADJACENT alla sola parte a contatto con la zona non termica
                old_area = s["area"]
                s["area"] = A_adj_dir

                # 2) crea una NUOVA superficie "opaque" esterna con area = diff
                new_surf = {
                    "name": f'{s.get("name","Adjacent split")} — external residual',
                    "type": "opaque",
                    "area": diff,
                    "sky_view_factor": s.get("sky_view_factor", 0.5),
                    "u_value": s.get("u_value", 1.4),
                    "solar_absorptance": s.get("solar_absorptance", 0.6),
                    "thermal_capacity": s.get("thermal_capacity", 0.0),
                    "orientation": {
                        "azimuth": az,
                        "tilt": tl,
                    },
                    "name_adj_zone": None,   # importantissimo: esterno, non adiacente
                }

                # se il coeff. conv./rad. sono già stati idratati, copiali (non obbligatorio)
                for k in (
                    "convective_heat_transfer_coefficient_internal",
                    "radiative_heat_transfer_coefficient_internal",
                    "convective_heat_transfer_coefficient_external",
                    "radiative_heat_transfer_coefficient_external",
                ):
                    if k in s: new_surf[k] = s[k]

                new_surfaces.append(new_surf)
                add_issue("INFO",
                        f"building_surface[{i}]",
                        f"Split {dir_code}: ridotta area ADJ da {old_area:.3f} a {A_adj_dir:.3f} m²; creata nuova OPAQUE esterna da {diff:.3f} m².",
                        fixed=True)

        # Aggiungo in coda TUTTE le nuove superfici
        if new_surfaces:
            bui_clean["building_surface"].extend(new_surfaces)


    return bui_clean, issues


    


# def sanitize_and_validate_BUI(bui: dict,
#                               fix: bool = False,
#                               eps: float = 1e-6,
#                               defaults: dict | None = None):
#     """
#     Validate and (optionally) sanitize a BUI.
    
#     Parameters
#     ----------
#     bui : dict
#         The BUI dictionary to validate.
#     fix : bool
#         If True, attempts to correct invalid/zero values (clipping to eps, defaults, etc.).
#     eps : float
#         Minimum value for clipping of positive quantities.
#     defaults : dict | None
#         Optional physical defaults, e.g.:
#         {
#           "opaque_u_default": 0.5,
#           "transparent_u_default": 1.6,
#           "g_default": 0.6
#         }
    
#     Returns
#     -------
#     bui_clean : dict
#         Copy of the BUI (modified if fix=True).
#     issues : list[dict]
#         List of issues found: {"level": "ERROR|WARN", "path": str, "msg": str, "fix_applied": bool}
#     """

#     bui_clean = deepcopy(bui)
#     issues = []
#     D = defaults or {
#         "opaque_u_default": 0.5,
#         "transparent_u_default": 1.6,
#         "g_default": 0.6
#     }

#     def add_issue(level, path, msg, fixed=False):
#         issues.append({"level": level, "path": path, "msg": msg, "fix_applied": fixed})

#     # ---------------------------
#     # 1) BUILDING-LEVEL CHECKS
#     # ---------------------------
#     b = bui_clean.get("building", {})
#     for key in ["net_floor_area", "exposed_perimeter", "height"]:
#         if key in b:
#             val = b[key]
#             if not (isinstance(val, (int, float)) and val > 0):
#                 if fix and isinstance(val, (int, float)) and val == 0:
#                     b[key] = max(eps, 1.0)
#                     add_issue("WARN", f"building.{key}", f"{key} era 0; impostato a {b[key]}", fixed=True)
#                 else:
#                     add_issue("ERROR", f"building.{key}", f"{key} deve essere > 0; valore={val}", fixed=False)

#     if "n_floors" in b and (not isinstance(b["n_floors"], (int, float)) or b["n_floors"] <= 0):
#         if fix and isinstance(b["n_floors"], (int, float)) and b["n_floors"] == 0:
#             b["n_floors"] = 1
#             add_issue("WARN", "building.n_floors", "n_floors era 0; impostato a 1", fixed=True)
#         else:
#             add_issue("ERROR", "building.n_floors", f"n_floors deve essere > 0; valore={b.get('n_floors')}", fixed=False)

#     # ---------------------------
#     # 2) SURFACES CHECKS
#     # ---------------------------
#     type_corrections = {"opque": "opaque", "opaqu": "opaque", "trasparent": "transparent"}
#     allowed_types = {"opaque", "transparent", "adiabatic", "adjacent"}

#     for i, s in enumerate(bui_clean.get("building_surface", [])):
#         path = f"building_surface[{i}]"
#         # Correggi/refusi type
#         t = s.get("type")
#         if isinstance(t, str) and t.lower() in type_corrections:
#             old = t
#             s["type"] = type_corrections[t.lower()]
#             add_issue("WARN", f"{path}.type", f'type="{old}" corretto in "{s["type"]}"', fixed=True)
#         t = (s.get("type") or "").lower()
#         if t not in allowed_types:
#             add_issue("ERROR", f"{path}.type", f'type "{s.get("type")}" non riconosciuto', fixed=False)

#         # area
#         area = s.get("area", None)
#         if not (isinstance(area, (int, float)) and area > 0):
#             if fix and isinstance(area, (int, float)) and area == 0:
#                 s["area"] = max(eps, 1.0)
#                 add_issue("WARN", f"{path}.area", f"area was 0; set to {s['area']}", fixed=True)
#             else:
#                 add_issue("ERROR", f"{path}.area", f"area must be > 0; value={area}", fixed=False)

#         # sky_view_factor
#         svf = s.get("sky_view_factor", None)
#         if not (isinstance(svf, (int, float)) and 0.0 <= svf <= 1.0):
#             if fix and isinstance(svf, (int, float)):
#                 s["sky_view_factor"] = min(1.0, max(0.0, svf))
#                 add_issue("WARN", f"{path}.sky_view_factor", f"sky_view_factor out of range; clippato a {s['sky_view_factor']}", fixed=True)
#             else:
#                 add_issue("WARN", f"{path}.sky_view_factor", f"sky_view_factor missing or not in [0,1]; value={svf}", fixed=False)

#         # U-value per opaque/transparent
#         if t in {"opaque", "transparent"}:
#             u = s.get("u_value", None)
#             if not (isinstance(u, (int, float)) and u > 0):
#                 if fix:
#                     default_u = D["opaque_u_default"] if t == "opaque" else D["transparent_u_default"]
#                     s["u_value"] = default_u
#                     add_issue("WARN", f"{path}.u_value", f"u_value was invalid ({u}); set to default {default_u}", fixed=True)
#                 else:
#                     add_issue("ERROR", f"{path}.u_value", f"u_value must be > 0; value={u}", fixed=False)

#         # g-value per trasparenti
#         if t == "transparent":
#             g = s.get("g_value", None)
#             if not (isinstance(g, (int, float)) and g > 0):
#                 if fix:
#                     s["g_value"] = D["g_default"]
#                     add_issue("WARN", f"{path}.g_value", f"g_value was invalid ({g}); set to default {D['g_default']}", fixed=True)
#                 else:
#                     add_issue("ERROR", f"{path}.g_value", f"g_value must be > 0; value={g}", fixed=False)

#             # dimensioni finestra
#             h = s.get("height", None)
#             w = s.get("width", None)
#             if not (isinstance(h, (int, float)) and h > 0):
#                 if fix:
#                     s["height"] = 1.0
#                     add_issue("WARN", f"{path}.height", f"height was invalid ({h}); set to 1.0", fixed=True)
#                 else:
#                     add_issue("ERROR", f"{path}.height", f"height must be > 0; value={h}", fixed=False)
#             if not (isinstance(w, (int, float)) and w > 0):
#                 if fix:
#                     s["width"] = 1.0
#                     add_issue("WARN", f"{path}.width", f"width was invalid ({w}); set to 1.0", fixed=True)
#                 else:
#                     add_issue("ERROR", f"{path}.width", f"width must be > 0; value={w}", fixed=False)

#             # parapet 0..height
#             para = s.get("parapet", None)
#             if para is not None and isinstance(para, (int, float)) and h is not None and isinstance(h, (int,float)):
#                 if not (0.0 <= para <= max(h, 0.0)):
#                     if fix:
#                         s["parapet"] = min(max(0.0, para if isinstance(para, (int,float)) else 0.0), h if isinstance(h,(int,float)) else 1.0)
#                         add_issue("WARN", f"{path}.parapet", f"parapet was out of range; clippato a {s['parapet']}", fixed=True)
#                     else:
#                         add_issue("WARN", f"{path}.parapet", f"parapet was out of range (0..height); value={para}, height={h}", fixed=False)

#         # orientamenti
#         ori = s.get("orientation", {})
#         tilt = ori.get("tilt", None)
#         az   = ori.get("azimuth", None)

#         # tilt 0 o 90 consigliati
#         if not (isinstance(tilt, (int, float)) and tilt in (0, 90)):
#             if fix and isinstance(tilt, (int, float)):
#                 s["orientation"]["tilt"] = 0 if abs(tilt - 0) < abs(tilt - 90) else 90
#                 add_issue("WARN", f"{path}.orientation.tilt", f"tilt was non standard ({tilt}); normalized to {s['orientation']['tilt']}", fixed=True)
#             else:
#                 add_issue("WARN", f"{path}.orientation.tilt", f"tilt was non standard ({tilt}); normalized to {s['orientation']['tilt']}", fixed=False)

#         # azimuth 0/90/180/270 consigliati
#         if not (isinstance(az, (int, float)) and az in (0, 90, 180, 270)):
#             if fix and isinstance(az, (int, float)):
#                 snapped = int(round(az / 90.0) * 90) % 360
#                 if snapped == 360: snapped = 0
#                 s["orientation"]["azimuth"] = snapped
#                 add_issue("WARN", f"{path}.orientation.azimuth", f"azimuth non standard ({az}); normalizzato a {snapped}", fixed=True)
#             else:
#                 add_issue("WARN", f"{path}.orientation.azimuth", f"azimuth consigliato in {{0,90,180,270}}; valore={az}", fixed=False)

#     # ---------------------------
#     # 3) ADJACENT ZONES CHECKS
#     # ---------------------------
#     for j, az in enumerate(bui_clean.get("adjacent_zones", [])):
#         path = f"adjacent_zones[{j}]"
#         vol = az.get("volume", None)
#         if not (isinstance(vol, (int, float)) and vol > 0):
#             if fix and isinstance(vol, (int, float)) and vol == 0:
#                 az["volume"] = 1.0
#                 add_issue("WARN", f"{path}.volume", "volume era 0; impostato a 1.0", fixed=True)
#             else:
#                 add_issue("ERROR", f"{path}.volume", f"volume deve essere > 0; valore={vol}", fixed=False)

#         # coerenza lunghezze
#         arrs = {
#             "area_facade_elements": az.get("area_facade_elements"),
#             "typology_elements": az.get("typology_elements"),
#             "transmittance_U_elements": az.get("transmittance_U_elements"),
#             "orientation_elements": az.get("orientation_elements"),
#         }
#         lengths = [len(v) for v in arrs.values() if isinstance(v, (list, np.ndarray))]
#         if lengths and len(set(lengths)) != 1:
#             add_issue("ERROR", path, f"Lengths of arrays are not coherent: {[ (k, None if v is None else len(v)) for k,v in arrs.items() ]}", fixed=False)

#         # U elementi > 0
#         U = arrs.get("transmittance_U_elements")
#         if isinstance(U, (list, np.ndarray)):
#             U = np.asarray(U, dtype=float)
#             bad = np.where(U <= 0)[0].tolist()
#             if bad:
#                 if fix:
#                     U[U <= 0] = D["opaque_u_default"]
#                     az["transmittance_U_elements"] = U
#                     add_issue("WARN", f"{path}.transmittance_U_elements", f"U<=0 ai idx {bad}; impostati a {D['opaque_u_default']}", fixed=True)
#                 else:
#                     add_issue("ERROR", f"{path}.transmittance_U_elements", f"U<=0 ai idx {bad}", fixed=False)

#     # -------------------------------------------------------------------
#     # 4) ADJACENCY PAIRING CAPS: OP (verticale) della zona ≤ area BUI
#     # -------------------------------------------------------------------
#     def _azimuth_to_dir(az: float) -> str | None:
#         """Mappa 0/90/180/270 -> 'NV'/'EV'/'SV'/'WV'. Altri valori -> None."""
#         if az is None:
#             return None
#         azn = int(round(float(az))) % 360
#         if azn in (0, 360): return "NV"
#         if azn == 90:       return "EV"
#         if azn == 180:      return "SV"
#         if azn == 270:      return "WV"
#         return None

#     # Mappa nome zona -> dizionario zona (per lookup rapido)
#     adj_by_name = { z.get("name"): z for z in bui_clean.get("adjacent_zones", []) }

#     for i, s in enumerate(bui_clean.get("building_surface", [])):
#         path = f"building_surface[{i}]"
#         if (s.get("type") != "adjacent") or (not s.get("name_adj_zone")):
#             continue

#         az = s.get("orientation", {}).get("azimuth", None)
#         tilt = s.get("orientation", {}).get("tilt", None)

#         # Consideriamo solo pareti verticali del BUI (tilt≈90) con direzione cardinale nota
#         if not (isinstance(tilt, (int,float)) and abs(tilt-90) < 1e-6):
#             continue

#         dir_card = _azimuth_to_dir(az)
#         if dir_card is None:  # orizzontali o azimut non standard -> skip
#             continue

#         A_bui = s.get("area", None)
#         if not (isinstance(A_bui, (int,float)) and A_bui > 0):
#             continue  # già segnalato sopra

#         adj_name = s["name_adj_zone"]
#         if adj_name not in adj_by_name:
#             add_issue("ERROR", path, f"name_adj_zone='{adj_name}' not found in adjacent_zones", fixed=False)
#             continue

#         azone = adj_by_name[adj_name]

#         # Assicuriamoci che gli array siano numpy *scrivibili*
#         A_arr  = np.array(azone.get("area_facade_elements", []), dtype=float).copy()
#         TypArr = np.array(azone.get("typology_elements", []),     dtype=object)
#         OriArr = np.array(azone.get("orientation_elements", []),  dtype=object)

#         if not (len(A_arr) == len(TypArr) == len(OriArr)):
#             add_issue("ERROR", f"adjacent_zones[{adj_name}]", "Arrays are not coherent (len mismatch) for the cap for direction", fixed=False)
#             continue

#         # Indici: OP and direction equal to dir_card (NV/EV/SV/WV)
#         mask = (np.char.upper(TypArr.astype(str)) == "OP") & (np.char.upper(OriArr.astype(str)) == dir_card)
#         idx = np.where(mask)[0]
#         if idx.size == 0:
#             # Nessuna parete OP in quella direzione: nulla da fare
#             continue

#         A_adj_dir = float(A_arr[idx].sum())
#         if A_adj_dir <= A_bui + 1e-9:
#             # OK: già ≤ BUI
#             continue

#         # Serve riduzione: o errore (fix=False) o correzione (fix=True)
#         msg = (f"Zona '{adj_name}' has OP area {dir_card}={A_adj_dir:.3f} m² superiore "
#                f"all’area BUI {A_bui:.3f} m² associated to {s.get('name','(senza nome)')}")
#         if not fix:
#             add_issue("ERROR", path, msg + " (no fix applied)", fixed=False)
#             continue

#         # Proportional reduction
#         sf = A_bui / max(A_adj_dir, eps)
#         A_arr[idx] = A_arr[idx] * sf

#         # Scriviamo indietro gli array *writable*
#         azone["area_facade_elements"] = A_arr
#         add_issue("WARN", path, msg + f" → redistributed proportionally (scale {sf:.3f})", fixed=True)

#     return bui_clean, issues
